'''
Riffle- "A segment of stream where the flow is shallower and more turbulent."

Riffle is concerned with abstracting persistent TCP/TLS connections. Exhaustively, 
riffle is a deferred RPC protocol over TLS based on twisted.pb.PerspectiveBroker that uses Avatars 
automagically created from self-signed SSL keys. 

Conceptually, there are three important parts:
    - interacting with remote clients through RPC 
    - managing persistent connections
    - connecting or listening for Riffle connections

More to follow. 
'''

from twisted.python.filepath import FilePath
from twisted.internet.endpoints import SSL4ServerEndpoint
from twisted.internet.task import react
from twisted.internet import reactor
from twisted.internet import defer
from twisted.spread import pb
from twisted.cred import portal

from zope.interface import implements

from twisted.internet.endpoints import SSL4ClientEndpoint
from twisted.internet.ssl import PrivateCertificate, Certificate, optionsForClientTLS


class Riffle(object):

    def __init__(self, host, port, secure=True):
        '''
        This is not a Twisted style 'protocol', its a wrapper around a modified PB. 
        The Riffle object makes creating new riffle sessions easier. 

        :param host: the hostname to connect/listen to 
        :type host: str.
        :param port: port to connect/listen to
        :type port: int
        :param secure: should riffle connect over TLS or not? This is a temporary thing
        :type secure: bool.
        '''
        self.host, self.port = host, port
        self.secure = secure

    def _serverContext(self, pem):
        ca = Certificate.loadPEM(pem)
        myCertificate = PrivateCertificate.loadPEM(pem)

        return SSL4ServerEndpoint(reactor, self.port, myCertificate.options(ca))

    def _clientContext(self, caCert, pkey):
        ctx = optionsForClientTLS(u"pds.production", Certificate.loadPEM(caCert), PrivateCertificate.loadPEM(pkey))
        return SSL4ClientEndpoint(reactor, self.host, self.port, ctx,)

    def serve(self, keypath):
        serverEndpoint = self._serverContext(keypath)
        serverEndpoint.listen(RiffleServerFactory(portal))

    @defer.inlineCallbacks
    def connect(self, caCert, keys):
        clientEndpoint = self._clientContext(caCert, keys)

        factory = RiffleClientFactory()
        clientEndpoint.connect(factory)

        avatar = yield factory.login(portal)

        defer.returnValue(Levy(avatar))


############################################################
# Realms, Portals, and Wrappers
# These classes are not meant to be subclassed
############################################################

class Portal(portal.Portal):

    '''
    Bi-directional portal. Stores connections and is responsible for allocating
    new connections with appropriate avatars.

    Currently takes a dict, but this isn't ideal
    '''

    def __init__(self):
        self.realms = {}

    def addRealm(self, matcher, realm):
        '''
        Add a realm to this portal with its corresponding matcher. This must be done
        before the server is started or connections are made, else requesting peers
        may fail to connect. 

        :param realm: the realm object that will handle the incoming connections
        :type realm: riffle.Realm
        :param matcher: an re matcher object that tests presented domain names for incoming clients
        :type matcher: re.matcher
        '''
        self.realms[matcher] = realm

    def findRealm(self, credential):
        '''
        Find the appropriate realm for the given credential. Matches are found using
        re filters. 
        '''
        # print self.realms
        # return self.realms[credential]

        for k, v in self.realms.iteritems():
            if k.match(credential):
                return v

        raise KeyError("No matcher was found to handle ", credential)

    def login(self, credentials, mind):
        print str(credentials) + ' attempting to login'
        target = self.findRealm(credentials)
        return target.requestAvatar(credentials, mind)

    def partialLogin(self, credentials):
        '''
        Clients send servers their representations before the server gives the client 
        its 'mind' object. This call returns the correct representation for the server
        *without* the mind object. Does not add the avatar to a realm. 
        '''
        target = self.findRealm(credentials)
        return target.requestPartialAvatar(credentials)


class Realm:

    '''
    Wraps a type of avatar and all connections for that avatar type
    '''
    implements(portal.IRealm)

    def __init__(self, avatar):
        self.avatar = avatar
        self.connections = set()

    @defer.inlineCallbacks
    def requestAvatar(self, avatarID, mind):
        '''
        Returns an instance of the appropriate avatar. Asks the avatar to perform 
        any needed initialization (which should be a deferred)
        '''
        avatar = self.avatar(avatarID, self)
        avatar.attached(mind)
        yield avatar.initialize()

        self.connections.add(avatar)

        # move detached from the avatar to here?
        defer.returnValue((avatar, lambda a=avatar: a.detached(mind)))

    def requestPartialAvatar(self, avatarID):
        return self.avatar(avatarID, self)

    def connectionClosed(self, avatar):
        ''' An avatar disconnected '''
        print 'Connection lost'
        self.connections.remove(avatar)


class Levy(object):

    ''' Wraps a remote object reference to allow __getattr__ magic '''

    def __init__(self, remote):
        self.remote = remote

    def __getattr__(self, item):
        def wrap(*args):
            return self.remote.callRemote(item, *args).addCallbacks(self.printValue, self.printError)

        return wrap

    def printValue(self, value):
        print repr(value)
        return value

    def printError(self, error):
        print 'Error Callback'
        print 'error', error


############################################################
# Avatar, Referencable, and Viewable base classes
############################################################

class RifflePerspective(pb.Avatar):

    def __init__(self, name, realm):
        self.name = name
        self.realm = realm

    @defer.inlineCallbacks
    def initialize(self):
        '''
        An initialization method that may hit the database or perform other model
        related tasks needed to initialize this avatar. This method is meant to be subclassed. 
        '''
        yield 1
        defer.returnValue(None)

    def attached(self, mind):
        self.remote = mind

    def detached(self, mind):
        self.remote = None
        self.realm.connectionClosed(self)


class RiffleReferencable(pb.Referenceable):

    def __init__(self, name, realm):
        self.name = name

    def connected(self, perspective):
        ''' Called when a remote user gains access to this object. Must save a reference '''
        self.perspective = perspective


class RiffleViewable(pb.Viewable):

    def view_doFoo(self, perspective, arg1, arg2):
        print 'Do Foo!', perspective, arg1, arg2
        return 'Done'


############################################################
# Perspective Broker Monkey Patches
############################################################

class RiffleClientFactory(pb.PBClientFactory):

    @defer.inlineCallbacks
    def login(self, portal):
        # Have to add connection to the portal
        self.portal = portal

        print 'Requesting root object'
        root = yield self.getRootObject()
        print 'Got root:', root

        peerCertificate = Certificate.peerFromTransport(self._broker.transport)
        pdid = peerCertificate.getSubject().commonName.decode('utf-8')

        # Returns the server's avatar based on the client's interpretation
        client = self.portal.partialLogin(pdid)
        client = pb.AsReferenceable(client, "perspective")

        avatar = yield root.callRemote('login', client)

        self.portal.login(pdid, avatar)

        defer.returnValue(avatar)


class RiffleServerFactory(pb.PBServerFactory):

    def __init__(self, portal):
        pb.PBServerFactory.__init__(self, portal)
        self.root = _RifflePortalRoot(portal)


class _RifflePortalRoot(pb._PortalRoot):

    def rootObject(self, broker):
        return _RifflePortalWrapper(self.portal, broker)


class _RifflePortalWrapper(pb._PortalWrapper):

    @defer.inlineCallbacks
    def remote_login(self, client):
        print 'Received a request for a remote login'

        peerCertificate = Certificate.peerFromTransport(self.broker.transport)
        pdid = peerCertificate.getSubject().commonName.decode('utf-8')
        print 'Got pdid', pdid

        avatar, logout = yield self.portal.login(pdid, client)
        avatar = pb.AsReferenceable(avatar, "perspective")

        # Formerly in _cbLogin, moved here to make the deferred chain cleaner
        puid = avatar.processUniqueID()

        # only call logout once, whether the connection is dropped (disconnect)
        # or a logout occurs (cleanup), and be careful to drop the reference to
        # it in either case
        logout = [logout]

        def maybeLogout():
            if not logout:
                return
            fn = logout[0]
            del logout[0]
            fn()

        self.broker._localCleanup[puid] = maybeLogout
        self.broker.notifyOnDisconnect(maybeLogout)

        defer.returnValue(avatar)


############################################################
# Utility Methods
############################################################

def dumpRealms(portal):
    for k, v in portal.realms.iteritems():

        for c in v.connections:
            print '\t', c

# Globally exposed portal object. Anyone using this class is going to need a portal,
# hence why its exposed

portal = Portal()



'''
Storage and stateful functionality for tools and paradrop. 

Holds onto the following things:
    Keys
    Identity
    Logs 
        Not really, Twisted is handling this-- we just make the directory here
'''

import os
import yaml

from os.path import expanduser
from pdtools.lib.output import out

# Check and see if we're on snappy or not. If not, then stick the logs in the local
# directory. We will be provisioned as a developer instance anyway, so the info doesn't matter yet,
# Also doesn't really matter for the buildtools, although it is a bit of a vulnurability
snappyPath = os.getenv("SNAP_APP_USER_DATA_PATH", None)

STORE_PATH = snappyPath + '/' if snappyPath is not None else expanduser('~') + '/.paradrop/'

LOG_PATH = STORE_PATH + 'logs/'
KEY_PATH = STORE_PATH + 'keys/'
MISC_PATH = STORE_PATH + 'misc/'
INFO_PATH = STORE_PATH + 'root.yaml'

# Keys required in the store
INFO_REQUIRES = ['version', 'pdid', 'chutes', 'routers', 'instances']


class Storage(object):

    def __init__(self, autoSave=True, router=False):
        '''
        Autosave saves baseConfig on change. (not implemented, see methods below)

        The router optional flag should be set to configure runtime store allocation specific 
        for VM routers
        '''
        self.autosave = autoSave

        # Check and see if we are running locally-- if so place the logs in a subdirectory
        # and respecify the paths
        if router:
            configureLocalPaths()

        # Create needed directories
        for path in [STORE_PATH, KEY_PATH, LOG_PATH, MISC_PATH]:
            if not os.path.exists(path):
                os.makedirs(path)

        # No config file found. Create a default one
        if not os.path.isfile(INFO_PATH):
            createDefaultInfo(INFO_PATH)

        self.baseConfig = loadYaml(INFO_PATH)

        # Sanity check contents of info and throw it out if bad
        if not validateInfo(self.baseConfig):
            out.warn('Saved configuration data invalid, destroying it.')
            os.remove(INFO_PATH)
            createDefaultInfo(INFO_PATH)
            self.baseConfig = loadYaml(INFO_PATH)

        # TODO: load keys

    def close(self):
        ''' Close and write out. This is automatic if autoSave is on'''
        pass

    def saveConfig(self, key, value):
        self.baseConfig[key] = value

        if self.autosave:
            writeYaml(self.baseConfig, INFO_PATH)

    def getConfig(self, key):
        return self.baseConfig[key]

    def saveKey(self, key, name):
        ''' Save the key with the given name. Overwrites by default '''
        path = KEY_PATH + name

        # Overwrite existing keys when asked for now
        # if os.path.isfile(path):
        #     print 'Key already exists!'
        #     pass

        with open(path, 'wb') as f:
            f.write(key)

    def getKey(self, name):
        ''' Returns the given key or None '''
        path = KEY_PATH + name

        if os.path.isfile(path):
            with open(path, 'rb') as f:
                return f.read()

        return None

    def saveMisc(self, contents, name):
        pass

    def loggedIn(self):
        ''' 
        Robustly determine if the account stored is logged in

        NOTE: not complete
        '''

        # currently just check for presence of saved id field
        return True if self.baseConfig['pdid'] else False


def configureLocalPaths():
    ''' Alter paths if the router is running as a vm '''
    global STORE_PATH, LOG_PATH, KEY_PATH, MISC_PATH, INFO_PATH

    # If not on snappy, we're a router running locally
    if not snappyPath:
        STORE_PATH = expanduser('~') + '/.paradrop/local/'

        LOG_PATH = STORE_PATH + 'logs/'
        KEY_PATH = STORE_PATH + 'keys/'
        MISC_PATH = STORE_PATH + 'misc/'
        INFO_PATH = STORE_PATH + 'root.yaml'


def validateInfo(contents):
    '''
    Error checking on the read YAML file. 

    :param contents: the read-in yaml to check
    :type contents: dict.
    :returns: True if valid, else false
    '''
    for k in INFO_REQUIRES:
        if k not in contents:
            return False

    # Check the validity of the contents

    return True


def createDefaultInfo(path):
    default = {
        'version': 1,
        'pdid': "",
        'chutes': [],
        'routers': [],
        'instances': []
    }

    writeYaml(default, path)


def writeYaml(contents, path):
    ''' Overwrites content with YAML representation at given path '''
    with open(path, 'w') as f:
        f.write(yaml.dump(contents, default_flow_style=False))


def loadYaml(path):
    ''' Return dict from YAML found at path '''
    with open(path, 'r') as f:
        return yaml.load(f.read())


# Global singleton (lite). This is NOT set up by default (else testing would be a
# nightmare.) The main script should set this up when ready to rock
store = Storage()



'''
Communication with routers
'''

import os
import yaml


from pdtools.coms import general
from pdtools.coms.client import RpcClient
from pdtools.lib.store import store

from twisted.internet import defer

from os.path import expanduser
from pdtools.lib.output import out

import requests
import json
import urllib

from pdtools.lib import pdutils
from pdtools.lib.store import store


@general.defaultCallbacks
@defer.inlineCallbacks
def provisionRouter(r, name, host, port):
    target = [x for x in store.getConfig('routers') if name in x]

    if len(target) == 0:
        print 'Router with name ' + name + ' not found.'
        yield 1
        defer.returnValue(None)

    client = RpcClient(host, port, '')
    ret = yield client.provision(target['_id'], target['publicKey'], target['privateKey'])


###############################################################################
# Chute Operations
###############################################################################

def installChute(host, port, config):
    '''
    Take a local config yaml file and launch chute of given host with pdfcd running on specified port.
    '''

    # Read local yaml into a config_json to send via post request
    try:
        with open(config, 'r') as stream:
            config_json = yaml.load(stream)
            config_json['date'] = config_json['date'].isoformat()
    except IOError as e:
        print 'Error: ' + config + ' not found.'
        return

    # Verify the config provided in some way.
    cfg_verf = pdutils.check(config_json, dict, {'dockerfile': dict, 'name': str, 'owner': str, 'host_config': dict})
    if cfg_verf:
        print 'ERROR: ' + cfg_verf
        return

    # Only allowed one way to provide a dockerfile order is local, remote, inline
    if 'local' in config_json['dockerfile']:
        with open(config_json['dockerfile']['local'], 'r') as stream:
            config_json['dockerfile'] = stream.read()
    elif 'remote' in config_json['dockerfile']:
        print 'Remote Dockerfile not supported yet.'
        return
    elif 'inline' in config_json['dockerfile']:
        config_json['dockerfile'] = config_json['dockerfile']['inline']
    else:
        print 'ERROR: No Dockerfile specified in config file.'
        return

    print 'Installing chute...'
    params = {'config': config_json}
    r = requests.post('http://' + host + ':' + str(port) + '/v1/chute/create', data=json.dumps(params), stream=True)
    for line in r.iter_lines():
        if line:
            try:
                line = json.loads(line)
                if line.get('success'):
                    print line.get('message')
                else:
                    print 'ERROR: Failed to install chute.(' + urllib.unquote(str(line.get('message'))) + ')'
            except Exception as e:
                print line


def deleteChute(host, port, name):
    '''
    Remove chute with given name from host with pdfcd running on specified port.
    '''

    print 'Removing chute...'

    params = {'name': name}
    r = requests.post('http://' + host + ':' + str(port) + '/v1/chute/delete', data=json.dumps(params))

    res = json.loads(r._content)
    if res.get('success'):
        print res.get('message')
    else:
        print 'ERROR: Failed to delete chute.(' + urllib.unquote(str(res.get('message'))) + ')'


def stopChute(host, port, name):
    '''
    Stop chute with given name from host with pdfcd running on specified port.
    '''

    print 'Stopping chute...'

    params = {'name': name}
    r = requests.post('http://' + host + ':' + str(port) + '/v1/chute/stop', data=json.dumps(params))

    res = json.loads(r._content)
    if res.get('success'):
        print res.get('message')
    else:
        print 'ERROR: Failed to stop chute.(' + urllib.unquote(str(res.get('message'))) + ')'


def startChute(host, port, name):
    '''
    Start chute with given name from host with pdfcd running on specified port.
    '''

    print 'Starting chute...'

    params = {'name': name}
    r = requests.post('http://' + host + ':' + str(port) + '/v1/chute/start', data=json.dumps(params))

    res = json.loads(r._content)
    if res.get('success'):
        print res.get('message')
    else:
        print 'ERROR: Failed to start chute.(' + urllib.unquote(str(res.get('message'))) + ')'



'''
Generic communication tasks. 

All remote calls are inline deferred callbacks. This allows them to be
easily chained together and fired asynchronously. 

Since calls are started with task.react, each method must take a reactor as its 
first parameter. 

Adding the defaultCallbacks decorator prints the result of the operation on 
success or failure. All failed calls return an exception which should be 
suitable for printing. 
'''

import re

from twisted.internet import defer

from pdtools.coms.client import RpcClient
from pdtools.lib.output import out
from pdtools.lib import pdutils
from pdtools.lib import store
from pdtools.lib import riffle

# SERVER_HOST = 'paradrop.io'
SERVER_HOST = 'localhost'
SERVER_PORT = 8015
SERVER_RIFFLE_PORT = 8016

###############################################################################
# Default Callbacks
###############################################################################


def failureCallbacks(f):
    def w(*args, **kwargs):
        return f(*args, **kwargs).addErrback(printFailure)

    return w


def defaultCallbacks(f):
    ''' 
    Wrap the API call in default handlers which print results. Optional 
    params are used to turn off specific handlers
    '''
    def w(*args, **kwargs):
        return f(*args, **kwargs).addCallbacks(printSuccess, printFailure)

    return w


def printSuccess(r):
    if isinstance(r, list):
        for x in r:
            print x
    else:
        print r


def printFailure(r):
    print 'Operation Failed!'
    print r


###############################################################################
# Generic Functionality
###############################################################################

@defaultCallbacks
@defer.inlineCallbacks
def logs(reactor, host, port):
    ''' Connect to the named device and ask for logs. '''
    client = RpcClient(host, port, 'internal')
    res = yield client.log()

    # comes back raw from the file, have to convert back to dics
    # converted = [pdutils.str2json(x) for x in res.strip().split('\n')]
    # for a in [out.messageToString(x) for x in converted]:
    #     print a

    # Cant do this more succintly until we find the encoding bugs
    for x in res.strip().split('\n'):
        try:
            j = pdutils.str2json(x)
            s = out.messageToString(j)
            print s
        except:
            print 'Malformed line, raw output: ' + x

    defer.returnValue('Done')


@defaultCallbacks
@defer.inlineCallbacks
def echo(reactor, host, port):
    clientKey = store.store.getKey('client.pem')
    caKey = store.store.getKey('ca.pem')

    riffle.portal.addRealm(re.compile(r'^pds.production$'), riffle.Realm(ServerPerspective))

    avatar = yield riffle.Riffle(host, int(port)).connect(caKey, clientKey)
    result = yield avatar.echo('Hello from a client!')
    defer.returnValue(result)


###############################################################################
# Riffle Convenience Methods
###############################################################################

class ServerPerspective(riffle.RifflePerspective):

    def perspective_echo(self, arg):
        print 'Client: server called echo'
        return arg


def connectServer():
    ''' Quick refactor method, returns a riffle client ready to communicate with the server '''
    riffle.portal.addRealm(u'pds.production', riffle.Realm(ServerPerspective))

    return riffle.Riffle('localhost', 4322).connect(store.KEY_PATH)



'''
Communication with the server
'''

from pdtools.coms import general
from pdtools.coms.client import RpcClient
from pdtools.lib.store import store

from twisted.internet import defer
import getpass

###############################################################################
# Authentication Callbacks
###############################################################################


def authCallbacks(f):
    def w(*args, **kwargs):
        return f(*args, **kwargs).addCallbacks(authSuccess, general.printFailure)

    return w


def authSuccess(r):
    store.saveConfig('pdid', r['_id'])
    store.saveKey(r['keys'], 'client.pem')
    store.saveKey(r['ca'], 'ca.pem')

    print 'You have been successfully logged in.'


@authCallbacks
@defer.inlineCallbacks
def login(reactor):
    name, password = None, None

    name = raw_input("Username: ")
    password = getpass.getpass()

    client = RpcClient(general.SERVER_HOST, general.SERVER_PORT, '')
    ret = yield client.login(name, password)
    defer.returnValue(ret)


@authCallbacks
@defer.inlineCallbacks
def register(reactor):
    name, email, password = raw_input("Username: "), raw_input("Email: "), getpass.getpass()

    client = RpcClient(general.SERVER_HOST, general.SERVER_PORT, '')
    ret = yield client.register(name, email, password)
    defer.returnValue(ret)


@general.failureCallbacks
@defer.inlineCallbacks
def ownership(reactor):
    ''' Get the list of owned resources from the server '''
    pdid = store.getConfig('pdid')
    client = RpcClient(general.SERVER_HOST, general.SERVER_PORT, '')
    ret = yield client.ownership(pdid)

    # Add these items to the store
    store.saveConfig('chutes', ret['chutes'])
    store.saveConfig('routers', ret['routers'])
    store.saveConfig('instances', ret['instances'])

    printOwned()

    defer.returnValue(ret)


@general.failureCallbacks
@defer.inlineCallbacks
def createRouter(reactor, name):
    ''' Get the list of owned resources from the server '''
    pdid = store.getConfig('pdid')

    client = RpcClient(general.SERVER_HOST, general.SERVER_PORT, '')
    ret = yield client.provisionRouter(pdid, name)

    print 'Router successfully created. Boot the router and run "provision," specifying the \
    address with port=14321'

    yield ownership(None)

    defer.returnValue(ret)


###############################################################################
# Utils
###############################################################################

# Show this individuallye (once for each type)
def printOwned():
    chutes = store.getConfig('chutes')
    routers = store.getConfig('routers')

    def pprint(d):
        print '\t' + d['_id']

    print 'routers'
    [pprint(x) for x in routers]

    print 'chutes'
    [pprint(x) for x in chutes]



'''
Router provisioning

NOTE: In writing this code Ive come to believe its easier and better
for us to host pre-installed and provisioned instances ourselves rather 
than build them from scratch.

We would host the image with all tools installed then provide it 
for download. How the provisioning process actually occurs I'm not sure. 
'''

import subprocess
import os
import wget


def provisionVirtual():
    '''
    Provision a virtual machine with snappy and paradrop.

        - Downloads the snappy vm and paradrop.snap, or takes a directory from which 
    to build paradrop.snap from. 
        - Mounts the VM and inserts keys and identifying information from the server. 
        - Unmount the image and remove the mounting directory  
    '''
    pass


def provisionHardware():
    '''
    Provision a physical machine with snappy and paradrop.

        - Downloads the snappy vm and paradrop.snap, or takes a directory from which 
    to build paradrop.snap from. 
        - Flashes snappy onto the target disk
        - Inserts keys and identifying information from the server. 
        - Unmount the image and remove the mounting directory  
    '''
    pass


def partedtest():

    path = os.path.dirname(os.getcwd()) + '/snappy-vm.img'

    # Get the offset
    output = subprocess.check_output("parted -s " + path + " unit B print", shell=True)
    target = [x for x in output.split('\n') if 'system-a' in x][0]

    # The second entry is the target offset, but this may not be kosher...
    offset = target.split()[1].replace('B', '')

    # Mount (this may not be necesary when provisioning flashed router)
    mountdir = '/mnt/snappytmp'
    output = subprocess.check_output("sudo mkdir " + mountdir, shell=True)
    output = subprocess.check_output("sudo mount -o loop,offset=%s %s %s" % (offset, path, mountdir), shell=True)


def cleanup():
    try:
        subprocess.check_output("sudo umount " + mountdir, shell=True)
    except:
        print 'Done'

###############################################################################
#  Downloading
###############################################################################


def downloadImage(defaultPath):
    '''
    Download a snappy image
    '''
    url = 'http://releases.ubuntu.com/15.04/ubuntu-15.04-snappy-amd64-generic.img.xz'

    path = raw_input("Enter a directory to save the image [" + defaultPath + ']') or defaultPath

    # Check and see if an image already exists
    # Ask the user if they want to save to a custom directory
    # Download the image



###################################################################
# Copyright 2013-2015 All Rights Reserved
# Authors: The Paradrop Team
###################################################################

"""
lib.utils.output.
Helper for formatting output from Paradrop.
"""

import time, json, urllib

timeflt = lambda: time.time()
timeint = lambda: int(time.time())
timestr = lambda x=None: time.asctime(time.localtime(x)) if x else time.asctime()

def timedur(x):
    """
        Print consistent string format of seconds passed.
        Example: 300 = '5 mins'
        Example: 86400 = '1 day'
        Example: 86705 = '1 day, 5 mins, 5 sec'
    """
    divs = [('days', 86400), ('hours', 3600), ('mins', 60)]
    x = float(x)
    res = []
    for lbl, sec in divs:
        if(x >= sec):
            rm, x = divmod(x, float(sec))
            # If exactly 1, remove plural of label
            if(rm == 1.0):
                res.append((lbl[:-1], int(rm)))
            else:
                res.append((lbl, int(rm)))
    
    # anything left over is seconds
    x = int(x)
    if(x == 1):
        res.append(("second", x))
    elif(x == 0):
        pass
    else:
        res.append(("seconds", x))
    
    return ", ".join(["%d %s" % (x[1], x[0]) for x in res])

def convertUnicode(elem):
    """Converts all unicode strings back into UTF-8 (str) so everything works.
        Call this function like:
            json.loads(s, object_hook=convertUnicode)"""
    if isinstance(elem, dict):
        return {convertUnicode(key): convertUnicode(value) for key, value in elem.iteritems()}
    elif isinstance(elem, list):
        return [convertUnicode(element) for element in elem]
    elif isinstance(elem, unicode):
        return elem.encode('utf-8')
    #DFW: Not sure if this has to be here, but deal with possible "null" MySQL strings
    elif(elem == 'null'):
        return None
    else:
        return elem

def urlEncodeMe(elem, safe=' '):
    """
        Converts any values that would cause JSON parsing to fail into URL percent encoding equivalents.
        This function can be used for any valid JSON type including str, dict, list.
        Returns:
            Same element properly encoded.
    """
    # What type am I?
    if isinstance(elem, dict):
        return {urlEncodeMe(key, safe): urlEncodeMe(value, safe) for key, value in elem.iteritems()}
    elif isinstance(elem, list):
        return [urlEncodeMe(element, safe) for element in elem]
    elif isinstance(elem, str):
        # Leave spaces alone, they are save to travel for JSON parsing
        return urllib.quote(elem, safe)
    else:
        return elem

def urlDecodeMe(elem):
    """
        Converts any values that would cause JSON parsing to fail into URL percent encoding equivalents.
        This function can be used for any valid JSON type including str, dict, list.
        Returns:
            Same element properly decoded.
    """
    # What type am I?
    if isinstance(elem, dict):
        return {urlDecodeMe(key): urlDecodeMe(value) for key, value in elem.iteritems()}
    elif isinstance(elem, list):
        return [urlDecodeMe(element) for element in elem]
    elif isinstance(elem, str):
        # Leave spaces alone, they are save to travel for JSON parsing
        return urllib.unquote(elem)
    else:
        return elem

def jsonPretty(j):
    """
        Returns a string of a JSON object in 'pretty print' format fully indented, and sorted.
    """
    return json.dumps(j, sort_keys=True, indent=4, separators=(',', ': '))

def json2str(j, safe=' '):
    """
        Properly converts and encodes all data related to the JSON object into a string format
        that can be transmitted through a network and stored properly in a database.
        Arguments:
            @j    : json to be converted
            @safe : optional, string of chars to pass to urlEncodeMe that are declared safe (don't encode)
    """
    return json.dumps(urlEncodeMe(j, safe), separators=(',', ':'))

def str2json(s):
    t = json.loads(s, object_hook=convertUnicode)
    # If t is a list, object_hook was never called (by design of json.loads)
    # deal with that situation here
    if(isinstance(t, list)):
        t = [convertUnicode(i) for i in t]
    # Make sure to still decode any strings
    return urlDecodeMe(t)

class dict2obj(object):
    def __init__(self, aDict=None, **kwargs):
        if(aDict is not None):
            aDict.update(kwargs)
        else:
            self.__dict__.update(kwargs)

def check(pkt, pktType, keyMatches=None, **valMatches):
    """This function takes an object that was expected to come from a packet (after it has been JSONized)
        and compares it against the arg requirements so you don't have to have 10 if() statements to look for keys in a dict, etc..
        
        Args:
            @pkt             : object to look at
            @pktType         : object type expected (dict, list, etc..)
            @keyMatches      : a list of minimum keys found in parent level of dict, expected to be an array
            @valMatches      : a dict of key:value pairs expected to be found in the parent level of dict
                              the value can be data (like 5) OR a type (like this value must be a @list@).
        Returns:
            None if everything matches, otherwise it returns a string as to why it failed."""
    #First check that the pkt type is equal to the input type
    if(type(pkt) is not pktType):
        return 'expected %s' % str(pktType)

    if(keyMatches):
        # Convert the keys to a set
        keyMatches = set(keyMatches)
        #The keyMatches is expected to be an array of the minimum keys we want to see in the pkt if the type is dict
        if(type(pkt) is dict):
            if(not keyMatches.issubset(pkt.keys())):
                return 'missing, "%s"' % ', '.join(list(keyMatches - set(pkt.keys())))
        else:
            return None

    #Finally for anything in the valMatches find those values
    if(valMatches):
        # Pull out the dict object from the "valMatches" key
        if('valMatches' in valMatches.keys()):
            matchObj = valMatches['valMatches']
        else:
            matchObj = valMatches

        for k, v in matchObj.iteritems():
            #Check for the key
            if(k not in pkt.keys()):
                return 'key missing "%s"' % k

            #See how we should be comparing it:
            if(type(v) is type):
                if(type(pkt[k]) is not v):
                    return 'key "%s", bad value type, "%s", expected "%s"' % (k, type(pkt[k]), v)

            else:
                #If key exists check value
                if(v != pkt[k]):
                    return 'key "%s", bad value data, "%s", expected "%s"' % (k, pkt[k], v)

    return None

def explode(pkt, *args):
    """This function takes a dict object and explodes it into the tuple requested.

        It returns None for any value it doesn't find.

        The only error it throws is if args is not defined.

        Example:
            pkt = {'a':0, 'b':1}
            0, 1, None = pdcomm.explode(pkt, 'a', 'b', 'c')
    """
    if(not args):
        raise PDError('EXPLODE', 'args must be provided')

    # If there is an error make sure to return a tuple of the proper length
    if(not isinstance(pkt, dict)):
        return tuple([None] * len(args))

    # Now just step through the args and pop off everything from the packet
    # If a key is missing, the pkt.get(a, None) returns None rather than raising an Exception
    return tuple([pkt.get(a, None) for a in args])




"""Paradrop build tools.

Usage:
    paradrop login
    paradrop register
    paradrop install <host> <port> <path-to-config> 
    paradrop delete <host> <port> <chute-name>
    paradrop start <host> <port> <chute-name>
    paradrop stop <host> <port> <chute-name>
    paradrop snap-install <host> <port>
    paradrop router-create <name>
    paradrop router-provision <name> <host> <port>
    paradrop list
    paradrop logs <host> <port>
    paradrop echo <host> <port>
    paradrop (-h | --help)
    paradrop --version
Options:
    -h --help     Show this screen.
    --version     Show version.
"""

from docopt import docopt
from twisted.internet import task

from pdtools.lib import output
from pdtools.coms import routers, general, server
from pdtools.lib.store import store


def main():
    # For now, don't grab STDIO and don't write random log noise to conosle
    output.out.startLogging(stealStdio=False, printToConsole=False)

    args = docopt(__doc__, version='Paradrop build tools v0.1')

    host, port = args['<host>'], args['<port>']

    if args['install']:
        return routers.installChute(args['<host>'], args['<port>'], args['<path-to-config>'])

    if args['delete']:
        return routers.deleteChute(args['<host>'], args['<port>'], args['<chute-name>'])

    if args['start']:
        return routers.startChute(args['<host>'], args['<port>'], args['<chute-name>'])

    if args['stop']:
        return routers.stopChute(args['<host>'], args['<port>'], args['<chute-name>'])

    if args['snap-install']:
        print 'Not implemented. Sorry, love.'

    if args['logs']:
        task.react(general.logs, (host, port))

    if args['echo']:
        task.react(general.echo, (host, port))

    if args['login']:
        task.react(server.login)

    if args['register']:
        task.react(server.register)

    # logged in calls
    if not store.loggedIn():
        print 'You must login first.'
        return

    if args['list']:
        task.react(server.ownership)

    if args['router-create']:
        task.react(server.createRouter, (args['<name>'],))

    if args['router-provision']:
        task.react(routers.provisionRouter, (args['<name>'], args['<host>'], args['<port>']))

    print 'Something broke'

if __name__ == '__main__':
    main()



'''
I know this is confusing naming, this is a temp fix
'''

from setuptools import setup, find_packages

setup(
    name="pdtools",
    version="0.1.39",
    author="Paradrop Labs",
    description="Paradrop deployment and build tools",
    install_requires=[
        'docopt',
        'requests',
        'twisted',
        'bcrypt',
        'service-identity',
        'colorama',
        'pyyaml'
    ],

    packages=find_packages(),

    entry_points={
        'console_scripts': [
            'paradrop=pdtools.main:main',
        ],
    },
)
